/*
 * Expressions evaluation program
 * Thomas Nemeth, 2003
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <getopt.h>
/* for robust_home() */
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>
/* */
#include <libepe/libepe.h>
#include "config.h"



char *robust_home() {
    if (getenv("HOME"))
        return getenv("HOME");
    else if (getenv("USER") && getpwnam(getenv("USER")))
        return getpwnam(getenv("USER"))->pw_dir;
    else if (getenv("LOGNAME") && getpwnam(getenv("LOGNAME")))
        return getpwnam(getenv("LOGNAME"))->pw_dir;
    else if ((getuid() != -1) && (getpwuid(getuid())))
        return getpwuid(getuid())->pw_dir;
    else
        return "/";
}


void ask_for_vars(TEnv *env) {
    char  cmd[MAXSTRLEN + 1];
    TTok *v = env->var_list->head;

    while (v) {
        if (! v->priority) {
            printf("Input variable [%s] value : ", v->token);
            memset(cmd, 0, MAXSTRLEN + 1);
            if ((! fgets(cmd, MAXSTRLEN, stdin)) && feof(stdin)) {
                printf("\n");
                return;
            }
            v->value = strtod(cmd, NULL);
        }
        v = v->next;
    }
}


void display_fctlist() {
    int i;

    printf(PROGRAM " functions list :\n");
    for (i = 1 ; func_tab[i].name ; i++) {
        printf(" %2d. %s(x)\n", i, func_tab[i].name);
    }
    printf("\n");
    exit(0);
}


void display_error(ulong code) {
    char errormessage[MAXSTRLEN + 1];

    get_errormsg(errormessage, MAXSTRLEN + 1, code);
    fprintf(stderr, "formula error : %s", errormessage);
    exit(-3);
}


void usage(int help) {
    printf(PROGRAM " v" VERSION ",  Thomas Nemeth 2003\n");
    if (help) {
        printf("Usage : " PROGRAM " [-h] [-l] [-v] [-c file] [-f functions]"
               " [-e expressions|-x name] [-w variables] [-p prec]\n"
               " -h            : this help\n"
               " -v            : program version\n"
               " -l            : display functions list\n"
               " -c file       : load configuration file\n"
               " -f functions  : load a list of functions\n"
               " -e expression : evaluate an expression\n"
               " -x name       : evaluate function named 'name'\n"
               " -w variables  : with variables values\n"
               " -p prec       : set precision\n"
               "\n");
    }
}


char* parse_argv(int argc, char *argv[], TEnv *env) {
    int   option;
    ulong err = 0;
    char *ev_fct = NULL;

    while ((option = getopt(argc, argv, "hvlc:f:e:w:x:p:")) != -1) {
        switch (option) {
            case 'h' :
                usage(1);
                exit(0);
                break;
            case 'v' :
                usage(0);
                exit(0);
                break;
            case 'l' :
                display_fctlist();
                break;
            case 'c' :
                err = load_config(optarg, env);
                if (err != 0) display_error(err);
                break;
            case 'f' :
                err = load_expressions(optarg, env);
                if (err != 0) display_error(err);
                break;
            case 'e' :
                err = load_expression("argument expr", NULL, optarg, env);
                if (err == 0)
                    ev_fct = xstrdup("argument expr");
                else
                    display_error(err);
                break;
            case 'w' :
                err = load_variables(optarg, env);
                if (err != 0) display_error(err);
                break;
            case 'x' :
                ev_fct = xstrdup(optarg);
                break;
            case 'p' :
                env->precision = atoi(optarg);
                break;
            default :
                usage(1);
                exit(-1);
                break;
        }
    }

    return ev_fct;
}


void display_tree(TEnv *env) {
    TFct *f = env->fct_list_head;
    TTok *v = env->var_list->head;
    TRng *r = env->rng_list_head;

    while (f) {
        TTok *e = f->head;
        while (e) {
            printf("%s(id%d)", e->token, e->id);
            e = e->next;
            if (e) printf(" ");
        }
        printf("\n");
        f = f->next;
    }
    while (v) {
        printf("%s(%f)", v->token, v->value);
        v = v->next;
        if (v) printf(" ");
    }
    while (r) {
        printf("%s in [%f,%f] @ %f %s\n",
                r->name, r->from, r->to, r->step,
                r->main ? "main" : "");
        r = r->next;
    }
    printf("\n");
}


int eval_input(TEnv *env) {
    char   cmd[MAXSTRLEN + 1];
    char   prec[MAXSTRLEN + 1];
    int    ret;
    double res;

    printf("Expression : ");
    memset(cmd, 0, MAXSTRLEN + 1);
    if ((! fgets(cmd, MAXSTRLEN, stdin)) && feof(stdin)) {
        printf("\n");
        return -2;
    }

    if ((ret = load_expression("input expr", NULL, cmd, env)) != 0) return ret;
    ask_for_vars(env);
    if ((ret = eval_function("input expr", env, &res)) != 0) return ret;

    snprintf(prec, MAXSTRLEN, " -> Result = %%.%df\n", env->precision);
    printf(prec, res);

    return 0;
}


int eval_fname(TEnv *env, const char *fname) {
    int    ret;
    char   prec[MAXSTRLEN + 1];

    if (env->rng_list_head) {
        TFct *res = NULL;
        TTok *r;
        if ((ret = eval_range(fname, env, &res)) != 0) return ret;
        r = res->head;
        snprintf(prec, MAXSTRLEN, "%%s  %%.%df\n", env->precision);
        while (r) {
            printf(prec, r->token, r->value);
            r = r->next;
        }
        free_list(&res);
    } else {
        double res;
        if ((ret = eval_function(fname, env, &res)) != 0) return ret;
        snprintf(prec, MAXSTRLEN, "%%.%df\n", env->precision);
        printf(prec, res);
    }

    return 0;
}


int main(int argc, char *argv[]) {
    TEnv  *env = init_env();
    int    ret = 0;
    char  *n = parse_argv(argc, argv, env);
    char configfile[MAXSTRLEN + 1];

    snprintf(configfile, MAXSTRLEN, "%s/%s", robust_home(), CONFIGFILE);
    load_config(configfile, env);
    if (! n)
        ret = eval_input(env);
    else {
        display_tree(env);
        ret = eval_fname(env, n);
        free(n);
    }

    free_env(&env);
    if (ret > 0) display_error(ret);

    return ret;
}
