/*
 * Expressions evaluation program
 * Thomas Nemeth, 2003
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#include "xmem.h"
#include "lists.h"
#include "parse.h"
#include "eval.h"



FctDef func_tab[] = {
    {"-",    opp,   NULL, 0.0},
    {"sqrt", sqrt,  "ge", 0.0},
    {"ln",   log,   "gt", 0.0},
    {"log",  log10, "gt", 0.0},
    {"e",    exp,   NULL, 0.0},
    {"sin",  sin,   NULL, 0.0},
    {"cos",  cos,   NULL, 0.0},
    {"tan",  tan,   NULL, 0.0},
    {"asin", asin,  NULL, 0.0},
    {"acos", acos,  NULL, 0.0},
    {"atan", atan,  NULL, 0.0}, /* atan2(x,y) ? */
    {NULL,   NULL,  NULL, 0.0}
};



double opp(double x) {
    return -x;
}


/* Returns
 * 0 : ok
 * 2 : unknown operator (should not happen)
 * 3 : division by 0
 */
int calc_op(double val1, double val2, TTok *ope, double *res) {
    int ret = 0;
    switch (ope->token[0]) {
        case '-': *res = val1 - val2;     break;
        case '+': *res = val1 + val2;     break;
        case '*': *res = val1 * val2;     break;
        case '/': if (val2 == 0) {
                      ret = DIVISIONBYZERO;
                      break;
                  }
                  *res = val1 / val2;
                  break;
        case '^': *res = pow(val1, val2); break;
        default:  ret = UNKNOWNOPER;
    }
#if DEBUG_LEVEL>0
    printf("Calculating : %.2f %s %.2f = %.2f\n", val1, ope->token, val2, *res);
#endif
    return ret;
}


/* BOOLEAN */
int good_range(int i, double v) {
    int ret = OK;

    if (func_tab[i].cmp) {
        char   *cmp = func_tab[i].cmp;
        double cval = func_tab[i].cval;
        if (strcmp(cmp, "ge") == 0)
            ret = (v >= cval);
        else if (strcmp(cmp, "gt") == 0)
            ret = (v > cval);
        else if (strcmp(cmp, "le") == 0)
            ret = (v <= cval);
        else if (strcmp(cmp, "lt") == 0)
            ret = (v < cval);
        else if (strcmp(cmp, "eq") == 0)
            ret = (v == cval);
    }
    return ret;
}


/* Returns
 * 0 : ok
 * 4 : out of range
 * 5 : unknown user function (should not happen)
 */
int calc_fct(double val, TTok *fct, TEnv *env, double *res) {
    int i, ret = UNKNOWNFUNC;

    if (fct->type == FUNCTION) {
        for (i = 0 ; func_tab[i].name ; i++) {
            if (strcmp(fct->token, func_tab[i].name) == 0) {
                if (! good_range(i, val)) {
                    ret = OUTOFRANGE;
                    break;
                }
                *res = func_tab[i].fct(val);
                ret = 0;
                break;
            }
        }
    } else {
        TFct *f = env->fct_list_head;
        while (f) {
            if (strcmp(fct->token, f->name) == 0) {
                char  *v = f->var;
                double x = find_var_value(env, v);
                if (v) set_var_value(env, v, val);
                ret = eval(f, env, res);
                if (v) set_var_value(env, v, x);
                break;
            }
            f = f->next;
        }
    }
#if DEBUG_LEVEL>0
    fprintf(stderr, "Calculating : %s(%.2f) = %.2f\n", fct->token, val, *res);
#endif
    return ret;
}


void set_fctvars_values(TFct *tree, TEnv *env) {
    TTok *v = env->var_list->head;

    while (v) {
        set_listvar_value(tree, v->token, v->value);
        v = v->next;
    }
}


/* Returns
 * 0 : ok
 * 1..n : error
 */
int eval(TFct *tree, TEnv *env, double *res) {
    TFct  *list = init_list(NULL);
    TTok   *c = tree->head;
    int     ret = 0;

    tree->eval = 1;
    set_fctvars_values(tree, env);
    while (c) {
        if ((c->type == VARIABLE) || (c->type == VALUE))
            push_copy(list, c);
        else if (c->type == OPERATOR) {
            double v2 = pop_value(list);
            double v1 = pop_value(list);
            double r;
            ret = calc_op(v1, v2, c, &r);
            if (ret == 0)
                push_res(list, r);
            else
                break;
        } else if (c->type == FUNCTION) {
            double v = pop_value(list);
            double r;
            ret = calc_fct(v, c, env, &r);
            if (ret == 0)
                push_res(list, r);
            else
                break;
        } else if (c->type == USERF) {
            TFct *f = env->fct_list_head;
            double v = 0.0;
            double r;
            while (f) {
                if (strcmp(c->token, f->name) == 0) {
                    if (f->var != NULL) v = pop_value(list);
                    break;
                }
                f = f->next;
            }
            ret = calc_fct(v, c, env, &r);
            if (ret == 0)
                push_res(list, r);
            else
                break;
        }
        c = c->next;
    }
    if (ret == 0)
        *res = list->head->value;
    else
        *res = 0.0;
    free_list(&list);
    tree->eval = 0;
    return ret;
}


int eval_function(const char *name, TEnv *env, double *res) {
    TFct  *fct = env->fct_list_head;
    TTok  *tok = xmalloc(sizeof(TTok));
    int    ret = 0;
    double val = 0.0;

    while (fct && (strcmp(fct->name, name) !=0)) fct = fct->next;
    if (fct) {
        tok->token = xstrdup(fct->name);
        tok->type  = USERF;
        val = find_var_value(env, fct->var ? fct->var : "x");
    } else {
        tok->token = xstrdup(name);
        tok->type  = FUNCTION;
        val = find_var_value(env, "x");
    }

    ret = calc_fct(val, tok, env, res);
    free(tok->token);
    return ret;
}


int eval_range(const char *name, TEnv *env, TFct **fct) {
    TRng  *rng = env->rng_list_head, *m = NULL;
    int    ret = 0;
    double val = 0.0;

    while (rng) {
        if (rng->main) m = rng;
        if ((rng->actv) && (!rng->eval)) break;
        rng = rng->next;
    }

    if (! rng) {
        double val, res;
        if (! m) m = env->rng_list_head;
        val = find_var_value(env, m->name);
        ret = eval_function(name, env, &res);
        if (! *fct) *fct = init_list(NULL);
        add_res(*fct, val, res, env);
        return ret;
    }

    rng->eval = 1;
    for (val = rng->from ; val <= rng->to ; val += rng->step) {
        set_var_value(env, rng->name, val);
        ret = eval_range(name, env, fct);
        if (ret != 0) break;
    }
    rng->eval = 0;

    return ret;
}
