/*
 * Expressions evaluation program
 * Thomas Nemeth, 2003
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include "xmem.h"
#include "lists.h"



TFct* init_list(char *expr) {
    TFct *l = xmalloc(sizeof(TFct));
    l->head  = NULL;
    l->queue = NULL;
    l->name  = NULL;
    l->expr  = expr;
    l->var   = NULL;
    l->eval  = 0;
    l->next  = NULL;

    return l;
}


TEnv* init_env() {
    TEnv *l = xmalloc(sizeof(TEnv));
    l->fct_list_head  = NULL;
    l->fct_list_queue = NULL;
    l->var_list       = init_list(NULL);
    l->rng_list_head  = NULL;
    l->rng_list_queue = NULL;
    l->precision      = 3;

    return l;
}


void add_tok(TFct *list, TTok *tok) {
    tok->next     = NULL;
    if (list->head == NULL)
        list->head = tok;
    else
        list->queue->next = tok;
    list->queue = tok;
}


void add_elt(TFct *list, const char *token, Type type, int priority, int id) {
    TTok *t = xmalloc(sizeof(TTok));

    t->token    = xstrdup(token);
    t->type     = type;
    if (type == VALUE)
        t->value = strtod(token, NULL);
    else
        t->value = 0.0;
    t->priority = priority;
    t->id       = id;
    add_tok(list, t);
}


void add_res(TFct *list, double var, double val, TEnv *env) {
    TTok *t = xmalloc(sizeof(TTok));
    char  p[MAXSTRLEN + 1];
    char  n[MAXSTRLEN + 1];

    snprintf(p, MAXSTRLEN, "%%.%df", env->precision);
    snprintf(n, MAXSTRLEN, p, var);
    t->token = xstrdup(n);
    t->value = val;
    add_tok(list, t);
}


void add_list(TFct *list, TFct *add) {
    if (list->head == NULL)
        list->head = add->head;
    else
        list->queue->next = add->head;
    list->queue = add->queue;
}


int exist_elt(TVar *list, const char *name) {
    TTok *v = list->head;

    while (v) {
        if (strcmp(v->token, name) == 0) return 1;
        v = v->next;
    }
    return 0;
}


void push_copy(TFct *list, TTok *t) {
    TTok *e = xmalloc(sizeof(TTok));

    e->token    = xstrdup(t->token);
    e->type     = t->type;
    e->priority = t->priority;
    e->id       = t->id;
    e->value    = t->value;
    e->next     = NULL;
    if (list->head == NULL)
        list->head = e;
    else
        list->queue->next = e;
    list->queue = e;
}


void push_res(TFct *list, double value) {
    TTok *e = xmalloc(sizeof(TTok));

    e->token    = NULL;
    e->type     = VALUE;
    e->priority = 0;
    e->id       = 0;
    e->value    = value;
    e->next     = NULL;
    if (list->head == NULL)
        list->head = e;
    else
        list->queue->next = e;
    list->queue = e;
}


TTok* pop_elt(TFct *list) {
    TTok *e = list->queue;
    TTok *p = list->head;

    if (!e) return NULL;
    if (p != e) {
        while (p->next != e) p = p->next;
        p->next = NULL;
        list->queue = p;
    } else {
        list->head = NULL;
        list->queue = NULL;
    }

    return e;
}


double pop_value(TFct *list) {
    TTok *e = pop_elt(list);
    double v = e->value;
    free_elt(&e);
    return v;
}


void pop_list(TFct *dest, TFct *src) {
    while (src->queue) {
        TTok *e = pop_elt(src);
#if DEBUG_LEVEL>0
        fprintf(stderr, "added %s\n", e->token);
#endif
        add_tok(dest, e);
    }
}


void add_var(TEnv *env, const char *name) {
    if (! exist_elt(env->var_list, name))
        add_elt(env->var_list, name, VARIABLE, 0, 0);
}


void set_listvar_value(TFct *list, const char *name, double value) {
    TTok *v = list->head;

    if (name) {
        while (v) {
            if ((v->type == VARIABLE) && (strcmp(v->token, name) == 0)) {
                v->value    = value;
                v->priority = 1;
            }
            v = v->next;
        }
    }
}


void set_var_value(TEnv *env, const char *name, double value) {
    set_listvar_value(env->var_list, name, value);
}


double find_var_value(TEnv *env, const char *name) {
    TTok *v = env->var_list->head;

    if (name) {
        while (v) {
            if (strcmp(v->token, name) == 0)
                return v->value;
            v = v->next;
        }
    }
    return 0.0;
}


void add_fct(TEnv *env, TFct *fct, const char *name, const char *var) {
    if (! env->fct_list_head) {
        env->fct_list_head  = fct;
        env->fct_list_queue = fct;
    } else {
        env->fct_list_queue->next = fct;
        env->fct_list_queue       = fct;
    }
    fct->name = xstrdup(name);
    fct->var  = xstrdup(var);
}


void add_rng(TEnv *env, const char *name, double from, double to, double step, double ismain) {
    TRng *rng = xmalloc(sizeof(TRng));

    rng->name = xstrdup(name);
    rng->from = from;
    rng->to   = to;
    rng->step = step;
    rng->main = ismain;
    rng->actv = 1;
    rng->eval = 0;
    rng->next = NULL;
    if (! env->rng_list_head) {
        env->rng_list_head  = rng;
        env->rng_list_queue = rng;
    } else {
        env->rng_list_queue->next = rng;
        env->rng_list_queue       = rng;
    }
}


void free_elt(TTok **elt) {
    FREE((*elt)->token);
    free(*elt);
    *elt = NULL;
}


void free_list(TFct **list) {
    TTok *t = (*list)->head;
    TTok *n = NULL;

    FREE((*list)->expr);
    while (t) {
        n = t->next;
        free_elt(&t);
        t = n;
    }
    free(*list);
    *list = NULL;
}


void free_env(TEnv **env) {
    TFct *f = (*env)->fct_list_head, *g;
    TRng *r = (*env)->rng_list_head, *s;

    while (f) {
        g = f->next;
        free_list(&f);
        f = g;
    }
    free_list(&((*env)->var_list));
    while (r) {
        s = r->next;
        FREE(r->name);
        r = s;
    }
    (*env)->fct_list_head  = NULL;
    (*env)->fct_list_queue = NULL;
    (*env)->rng_list_head  = NULL;
    (*env)->rng_list_queue = NULL;
    *env = NULL;
}
