/*
 * Expressions evaluation program
 * Thomas Nemeth, 2003
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#include "xmem.h"
#include "lists.h"
#include "parse.h"
#include "load.h"


int split_function(char *func, char **name, char **var, char **expr) {
    int   ret = 0;
    char *ftok, *fptr;

    if ((ftok = strtok_r(func, "=", &fptr)) == NULL) ret = READERROR;
    if (strchr(ftok, '(') != NULL) {
        char *dtok, *dptr;
        char *dcl = xstrdup(ftok);
        if ((dtok = strtok_r(dcl, "(", &dptr)) == NULL) ret = READERROR;
        *name = xstrdup(dtok);
        if ((dtok = strtok_r(dcl, ")", &dptr)) == NULL) ret = READERROR;
        *var = xstrdup(dtok);
        FREE(dcl);
    } else {
        *name = xstrdup(ftok);
        *var = NULL;
    }
    if ((ftok = strtok_r(NULL, "", &fptr)) == NULL) ret = READERROR;
    *expr = xstrdup(ftok);

    return ret;
}


ulong load_expression(const char *name, const char *var, const char *expr, TEnv *env) {
    TFct *fct;
    int   ret;

    ret = load(expr, env, &fct);
    if (ret != 0) {
        free_list(&fct);
        return ret;
    }
    add_fct(env, fct, name, var);

    return 0;
}


ulong load_expressions(const char *exp_list, TEnv *env) {
    int   ret = 0;
    int   fnum = 0;
    char *list = xstrdup(exp_list);
    char *ltok, *lptr;

    ltok = strtok_r(list, ";", &lptr);
    while (ltok) {
        char *func = xstrdup(ltok);
        char *name = NULL, *var = NULL, *expr = NULL;
        fnum++;
        if ((ret = split_function(func, &name, &var, &expr)) == 0)
            ret = load_expression(name, var, expr, env);
        FREE(name);
        FREE(expr);
        FREE(var);
        FREE(func);
        if (ret != 0) break;
        ltok = strtok_r(NULL, ";", &lptr);
    }

    FREE(list);
    if (ret > 0) ret += FCTLISTERR * fnum;
    return ret;
}


int load_variable(char *var, TEnv *env) {
    int ret = 0;
    char *ftok, *fptr;
    char *name = NULL;
    double value = 0.0;

    if ((ftok = strtok_r(var, "=", &fptr)) == NULL) ret = VARLISTERR;
    name = xstrdup(ftok);
    if ((ftok = strtok_r(NULL, "", &fptr)) == NULL) ret = VARLISTERR;
    if (ret == 0) {
        value = strtod(ftok, NULL);
        add_var(env, name);
        set_var_value(env, name, value);
    }
    FREE(name);

    return ret;
}


int load_range(char *var, TEnv *env) {
    int    ret = 0, ismain = 0, div = 0;
    char   format[MAXSTRLEN + 1];
    char   name[MAXSTRLEN + 1];
    double from, to, step = 1.0;

    strcpy(format, "%s in [%lf,%lf]");

    if (strchr(var, '@') != NULL) {
        div = 1;
        strcat(format, " @ %lf");
    } else if (strchr(var, '/') != NULL) {
        div = 2;
        strcat(format, " / %lf");
    }
    
    if (strstr(var, " main") != NULL) {
        ismain = 1;
        strcat(format, " main");
    }

    if (div != 0)
        sscanf(var, format, &name, &from, &to, &step);
    else
        sscanf(var, format, &name, &from, &to);
 
    if (div == 2) step = (to - from) / step;

    if (ret == 0) {
        add_var(env, name);
        add_rng(env, name, from, to, step, ismain);
    }

    return ret;
}


ulong load_variables(const char *var_list, TEnv *env) {
    int   ret = 0;
    int   vnum = 0;
    char *list = xstrdup(var_list);
    char *ltok, *lptr;

    ltok = strtok_r(list, ";", &lptr);
    while (ltok) {
        char *var = xstrdup(ltok);
        vnum++;
        if (strchr(var, '=') != NULL)
            ret = load_variable(var, env);
        else
            ret = load_range(var, env);
        FREE(var);
        if (ret != 0) break;
        ltok = strtok_r(NULL, ";", &lptr);
    }

    FREE(list);
    if (ret > 0) ret += vnum;
    return ret;
}


ulong load_config(const char *filename, TEnv *env) {
    FILE *file;
    int   ret = 0;
    int   lnum = 0;

    if ((file = fopen(filename, "r")) == NULL) return UNKNOWNFILE;

    while (! feof(file)) {
        char line[MAXSTRLEN + 1];
        char *etok, *eptr, *type, *data;
        memset(line, 0, MAXSTRLEN + 1);
        fgets(line, MAXSTRLEN, file);
        if ((line[0] == 0) || (line[0] == '#')) continue;
        lnum++;
        if ((etok = strtok_r(line, ":", &eptr)) == NULL) ret = READERROR;
        type = xstrdup(etok);
        if ((etok = strtok_r(NULL, "", &eptr)) == NULL) ret = READERROR;
        data = xstrdup(etok);
        if (ret == 0) {
            if (strcmp(type, "F") == 0)
                ret = load_expressions(data, env);
            else if (strcmp(type, "V") == 0) {
                ret = load_variables(data, env);
            } else {
                ret = UNKNOWNENTRY;
            }
        }
        FREE(type);
        FREE(data);
        if (ret != 0) break;
    }
    fclose(file);
    if (ret > 0) ret += CFGFILEERR * lnum;
    return ret;
}


void get_errormsg(char *msgstr, int maxlen, ulong errorcode) {
    memset(msgstr, 0, maxlen);
    while (errorcode > 0) {
        long val = 0;
        if (errorcode > CFGFILEERR) {
            ulong l = errorcode / CFGFILEERR;
            char  s[MAXSTRLEN + 1];
            sprintf(s, "Error in file at line %ld.\n", l);
            strcatn(msgstr, maxlen, s);
            val = l * CFGFILEERR;
        } else if (errorcode > FCTLISTERR) {
            ulong f = errorcode / FCTLISTERR;
            char  s[MAXSTRLEN + 1];
            sprintf(s, "Error at function #%ld.\n", f);
            strcatn(msgstr, maxlen, s);
            val = f * FCTLISTERR;
        } else if (errorcode > VARLISTERR) {
            ulong v = errorcode - VARLISTERR;
            char  s[MAXSTRLEN + 1];
            sprintf(s, "Error at variable #%ld.\n", v);
            strcatn(msgstr, maxlen, s);
            val = errorcode;
        } else if (errorcode > PARSEERROR) {
            ulong c = errorcode - PARSEERROR;
            char  s[MAXSTRLEN + 1];
            sprintf(s, "Parse error at character %ld.\n", c);
            strcatn(msgstr, maxlen, s);
            val = errorcode;
        } else {
            char *str;
            switch (errorcode) {
                case 1 : str = "Check error.\n";        break;
                case 2 : str = "Unknown operator.\n";   break;
                case 3 : str = "Division by zero.\n";   break;
                case 4 : str = "Value out of range.\n"; break;
                case 5 : str = "Unknown function.\n";   break;
                case 6 : str = "Unknown file.\n";       break;
                case 7 : str = "Unknown entry.\n";      break;
                case 8 : str = "Read error.\n";         break;
                default: str = "Unknown error.\n";      break;
            }
            strcatn(msgstr, maxlen, str);
            val = errorcode;
        }
        errorcode -= val;
    }
        
}
