/*
 * Expressions evaluation program
 * Thomas Nemeth, 2003
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include "xmem.h"
#include "lists.h"
#include "eval.h"
#include "parse.h"



Type add_token(char *tok, State s, TFct *ops, TFct *tree, int pos, TEnv *env) {
    static int idnum = 0;
    Type t = NOTHING;
    int  p = 0;

    if (tok[0] == 0) return t;
    if (s == S_VALU)
        t = VALUE;
    else if (s == S_OPER) {
        t = OPERATOR;
        switch (tok[0]) {
            case '+': p = 1; break;
            case '-': p = 2; break;
            case '*': p = 3;
            case '/': p = 4; break;
            case '^': p = 5; break;
            default:  p = 0; break;
        }
    } else {
        int i;
#if DEBUG_LEVEL>0
        printf("position %d -> %c\n", pos, tree->expr[pos]);
#endif
        for (i = 0; func_tab[i].name ; i++) {
            if (strcmp(tok, func_tab[i].name) == 0) {
                t = FUNCTION;
                p = 5;
                break;
            }
        }
        if (t == NOTHING) {
            TFct *f = env->fct_list_head;
            while (f) {
                if (strcmp(tok, f->name) == 0) {
                    t = USERF;
                    p = 5;
                    break;
                }
                f = f->next;
            }
            if ((tree->expr[pos] != '(') &&
                (((t == USERF) && (f->var != NULL)) ||
                 ((t == FUNCTION) && (tok[0] != '-')))) return NOTHING;
        }
        if (t == NOTHING) {
            t = VARIABLE;
            add_var(env, tok);
        }
    }
    idnum++;
    if ((t == VALUE) || (t == VARIABLE)) {
#if DEBUG_LEVEL>0
        fprintf(stderr, "added %s\n", tok);
#endif
        add_elt(tree, tok, t, p, idnum);
    } else {
        while ((ops->queue) && (p < ops->queue->priority)) {
            TTok *e = pop_elt(ops);
#if DEBUG_LEVEL>0
            fprintf(stderr, "added %s\n", e->token);
#endif
            add_tok(tree, e);
        }
        add_elt(ops, tok, t, p, idnum);
    }
    return t;
}


int parse(TFct *tree, TEnv *env, int *i) {
    int   j = *i, k = 0, ret = OK;
    char  token[MAXSTRLEN];
    Type  t = NOTHING;
    State state = S_NONE, oldstate = S_NONE;
    TFct *op_list = init_list(NULL);

    while (j < strlen(tree->expr)) {
        char c = tree->expr[j++];
        if (((c >= '0') && (c <= '9')) || (c == '.'))
            state = S_VALU;
        else if ((c=='+') || (c=='-') || (c=='*') || (c=='/') || (c == '^'))
            state = S_OPER;
        else {
            switch (c) {
                case '(': state = S_OPAR; break;
                case ')': state = S_CPAR; break;
                case ' ': state = S_NONE; break;
                default:  state = S_IDEN; break;
            }
        }
        if ((state == oldstate) && (state == S_OPER)) {
            if (c == '-')
                state = S_IDEN;
            else {
                ret = ERR;
                break;
            }
        }
        if (oldstate == S_NONE) oldstate = state;
        if (state != oldstate) {
#if DEBUG_LEVEL>0
            fprintf(stderr, "found %s of type %d ending at %d\n",
                    token, oldstate, j-1);
#endif
            if ((oldstate != S_OPAR) && (oldstate != S_CPAR)) {
                t = add_token(token, oldstate, op_list, tree, j - 1, env);
                if (t == NOTHING) {
                    ret = ERR;
                    break;
                }
            }
            oldstate = state;
            k = 0;
        }
        if ((state != S_OPAR) && (state != S_CPAR)) {
            token[k] = c;
            token[++k] = 0;
        } else if (state == S_OPAR) {
            if ((t == VALUE) || (t == VARIABLE)) {
                ret = ERR;
                break;
            }
            ret = parse(tree, env, &j);
            token[0] = 0;
            if (! ret) break;
        } else if (state == S_CPAR) {
            break;
        }
    }
    pop_list(tree, op_list);
    free_list(&op_list);
    *i = j;
    return ret;
}


char* normalize(const char *expr) {
    char *text = xmalloc(strlen(expr) + 1);
    int   i, j = 0;

    for (i = 0; expr[i] ; i++) {
        if ((expr[i] != ' ') && (expr[i] != '\n')) text[j++] = expr[i];
    }
    text[j++] = ' ';
    text[j] = 0;
#if DEBUG_LEVEL>0
    fprintf(stderr, "Normalizing done...\n");
#endif

    return text;
}


int check(const char *expr) {
    int i, nbp = 0;

    for (i = 0 ; i <= strlen(expr) ; i++) {
        if (expr[i] == '(')
            nbp++;
        else if (expr[i] == ')')
            nbp--;
    }
#if DEBUG_LEVEL>0
    fprintf(stderr, "Check %s...\n", nbp == 0 ? "ok" : "failed");
#endif

    return (nbp == 0);
}


/* Returns :
 * 0 : ok
 * 1 : check error
 * >=10 : parse error
 */
int load(const char *expr, TEnv *env, TFct **fct) {
    int    i = 0, ret = 0;
    TFct *fct_tree = init_list(normalize(expr));

#if DEBUG_LEVEL>0
    fprintf(stderr, "Expression : [%s]\n", fct_tree->expr);
#endif
    if (! check(fct_tree->expr)) {
        ret = CHECKERROR;
    } else if (! parse(fct_tree, env, &i)) {
#if DEBUG_LEVEL>0
        fprintf(stderr, "Parse error at character %c\n", fct_tree->expr[i]);
#endif
        ret = PARSEERROR + i;
    }

    *fct = fct_tree;
    return ret;
}

